#include "MappingResults.h"
#include <string>
#include <vector>
#include <algorithm>
#include <fstream>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <set>
using namespace std;

bool enrichedPairSortCriterion(const enrichedPair& ep1, const enrichedPair& ep2)
{
   return ep1.pValue < ep2.pValue;
}

bool compareOntoFin(const pair<string, double>& firstP, const pair<string, double>& secondP)
{
   return firstP.second>secondP.second;
}

const double PValueCutoff=0.001;

MappingResults::MappingResults(string mName, vector<string>& mMapped)
{
   name=mName;
   mapped=mMapped;
   uniqueAndOrder();
}

MappingResults::MappingResults(char* fileName)
{
   ifstream in(fileName);
   string line;
   getline(in, line);
   if (line[0]!='>')
   {
      cout<<"Not fasta format, program exit"<<endl;
      exit(0);
   }
   name=line.substr(1);
   double pValue;
   while (getline(in, line))
   {
      enrichedPair ep;
      istringstream is(line);
      is>>ep.name>>ep.pValue;
      enriched.push_back(ep);
   }
     
   calculateAdjustedPValue();         
}

MappingResults::~MappingResults()
{}

int MappingResults::getCommonEleCount(vector<string>& theOtherSet)
{
   vector<string> tmp;
   set_intersection(mapped.begin(), 
                    mapped.end(),  
                    theOtherSet.begin(), 
                    theOtherSet.end(), 
                    back_inserter(tmp));
   return tmp.size();
}

void MappingResults::uniqueAndOrder()
{
   sort(mapped.begin(), mapped.end());
   unique(mapped.begin(), mapped.end());
}

void MappingResults::addElement(string anEle)
{
   mapped.push_back(anEle);
}

void MappingResults::printFasta(ostream &out)
{
   out<<">"<<name<<endl;
   vector<string>::iterator vItr;
   for (vItr=mapped.begin(); vItr!=mapped.end(); ++vItr)
      out<<*vItr<<endl;
} 

void MappingResults::printEnriched(ostream &out, double cutoffPValue)
{
   sort(enriched.begin(), enriched.end(), enrichedPairSortCriterion);
   vector<enrichedPair>::iterator mItr;
   out<<">"<<name<<endl;
   for (mItr=enriched.begin(); mItr!=enriched.end(); ++mItr)
      if (mItr->pValue<=cutoffPValue)
      {
         out<<mItr->name<<"\t"
            <<std::scientific<<std::showpoint<<std::setprecision(8)
            <<mItr->pValue<<"\t"<<mItr->adjustedPValue<<endl;
      }
} 

void MappingResults::addEnriched(string termName, double pValue)
{
   enrichedPair ep;
   ep.name=termName;
   ep.pValue=pValue;
   enriched.push_back(ep);
} 

void MappingResults::addEnriched(string termName, double pValue, double logLikelihood, double uncertaintyCoefficient)
{
   enrichedPair ep;
   ep.name=termName;
   ep.pValue=pValue;
   ep.LogLikelihood=logLikelihood;
   ep.UncertaintyCoefficient=uncertaintyCoefficient;
   enriched.push_back(ep);
}


void MappingResults::calculateAdjustedPValue()
{
   sort(enriched.begin(), enriched.end(), enrichedPairSortCriterion);
   int i, j=enriched.size();
   double k;
   map<double, int> tieCount;
   for (i=0; i<j; ++i)
   {
      if (i==0)
         tieCount[enriched[i].pValue]=0;
      else
         if (tieCount[enriched[i].pValue]==0)
            tieCount[enriched[i].pValue]=i;
   }

   for (i=0; i<j; ++i)
   {
      k=enriched[i].pValue*j/(j-tieCount[enriched[i].pValue]);
      if (k>=1)
         enriched[i].adjustedPValue=1;
      else
         enriched[i].adjustedPValue=k; 
   }
}

void MappingResults::calculateSummaryScore(MappingResults& m)
{
   double d = calculateSMS(m.getEnriched());
   summaryScore[m.getName()]=d;
}

double MappingResults::calculateSMS(vector<enrichedPair> &en)
{
   int i, j=en.size()+1, comm=0, k;
   double prd=0;
   map<string, int> locMap;

   for (i=1; i<j; ++i)
   {
      if (en[i-1].adjustedPValue<PValueCutoff)
         locMap[en[i-1].name]=i;
   }
   
   k=locMap.size();

   for (j=0; j<enriched.size(); ++j)
   {
      if (locMap[enriched[j].name]>0 && enriched[j].adjustedPValue < PValueCutoff)
      {
         ++comm;
         double adj1=en[locMap[enriched[j].name]-1].adjustedPValue;
         double adj2=enriched[j].adjustedPValue;
         if (adj1==0)
           adj1=log(1e-323);
         else
           adj1=log(adj1);

         if (adj2==0)
           adj2=log(1e-323);
         else
           adj2=log(adj2);

         prd += adj1*adj2;
      }
   }

   if (k==comm)
      return prd;
   else
      return (prd/(k-comm));       
}

void MappingResults::calculateSummaryScore(MappingResults& m, double cutoff, int percent) 
{
   double d = calculateSMS(m.getEnriched(), cutoff);
   summaryScoreWithCutOff[m.getName()][percent]=d;
}

void MappingResults::calculateSummaryScore(MappingResults& m, double cutoff)
{
   double d = calculateSMS(m.getEnriched(), cutoff);
   summaryScoreWithCutOff[m.getName()][cutoff]=d;
}

void MappingResults::calculateSummaryScoreForGene(MappingResults& m, double cutoff)
{
   double d = calculateSMSForGene(m.getEnriched(), cutoff);
   summaryScoreWithCutOff[m.getName()][cutoff]=d;
}

void MappingResults::calculateSummaryScoreForGeneNoPenalty(MappingResults& m, double cutoff)
{
   double d = calculateSMSForGeneNoPenalty(m.getEnriched(), cutoff);
   summaryScoreWithCutOff[m.getName()][cutoff]=d;
}

double MappingResults::calculateSMS(vector<enrichedPair> &en, double cutoff)
{
   int i, j=en.size()+1, comm=0, k;
   double prd=0;
   map<string, int> locMap;

   for (i=1; i<j; ++i)
   {
      if (en[i-1].adjustedPValue<cutoff)
         locMap[en[i-1].name]=i;
   }

   k=locMap.size();

   for (j=0; j<enriched.size(); ++j)
   {
      if (locMap[enriched[j].name]>0 && enriched[j].adjustedPValue < cutoff )
      {
         ++comm;
         double adj1=en[locMap[enriched[j].name]-1].adjustedPValue;
         double adj2=enriched[j].adjustedPValue;
         if (adj1==0)
           adj1=log(1e-323);
         else
           adj1=log(adj1);

         if (adj2==0)
           adj2=log(1e-323);
         else
           adj2=log(adj2);

         prd += adj1*adj2;
      }
   }

   if (k==comm)
      return prd;
   else
      return (prd/(k-comm));
}

double MappingResults::calculateSMSForGene(vector<enrichedPair> &en, double cutoff)
{
   int i, j=en.size()+1, comm=0, k;
   double prd=0;
   map<string, int> locMap;

   for (i=1; i<j; ++i)
   {
      if (en[i-1].adjustedPValue<cutoff)
         locMap[en[i-1].name]=i;
   }

   k=locMap.size();

   int x=0;
   for (j=0; j<enriched.size(); ++j)
   {
      if (enriched[j].adjustedPValue<cutoff)
         ++x;

      if (locMap[enriched[j].name]>0 && enriched[j].adjustedPValue < cutoff )
      {
         ++comm;
         double adj1=en[locMap[enriched[j].name]-1].adjustedPValue;
         double adj2=enriched[j].adjustedPValue;
         if (adj1==0)
           adj1=log(1e-323);
         else
           adj1=log(adj1);

         if (adj2==0)
           adj2=log(1e-323);
         else
           adj2=log(adj2);

         prd += adj1*adj2;
      }
   }

   if ( (k+x-2*comm)<2 )
      return prd;
   else
      return (2*prd/(k+x-2*comm));
}

double MappingResults::calculateSMSForGeneNoPenalty(vector<enrichedPair> &en, double cutoff)
{
   int i, j=en.size()+1, comm=0, k;
   double prd=0;
   map<string, int> locMap;

   for (i=1; i<j; ++i)
   {
      if (en[i-1].adjustedPValue<cutoff)
         locMap[en[i-1].name]=i;
   }

   k=locMap.size();

   int x=0;
   for (j=0; j<enriched.size(); ++j)
   {
      if (enriched[j].adjustedPValue<cutoff)
         ++x;

      if (locMap[enriched[j].name]>0 && enriched[j].adjustedPValue < cutoff )
      {
         ++comm;
         double adj1=en[locMap[enriched[j].name]-1].adjustedPValue;
         double adj2=enriched[j].adjustedPValue;
         if (adj1==0)
           adj1=log(1e-323);
         else
           adj1=log(adj1);

         if (adj2==0)
           adj2=log(1e-323);
         else
           adj2=log(adj2);

         prd += adj1*adj2;
      }
   }

   return prd;
}

void MappingResults::calculateTopOntoPairs(MappingResults &m, string parentChildFile, double cutoff, int numbTopOnto)
{
   vector<pair<string, double> > ontoFin;
   vector<enrichedPair> en=m.getEnriched();
   int i, j=en.size()+1, comm=0, k;
   double prd=0;
   map<string, int> locMap;

   for (i=1; i<j; ++i)
   {
      if (en[i-1].adjustedPValue<cutoff)
         locMap[en[i-1].name]=i;
   }

   k=locMap.size();

   for (j=0; j<enriched.size(); ++j)
   {
      if (locMap[enriched[j].name]>0 && enriched[j].adjustedPValue < cutoff )
      {
         ++comm;
         double adj1=en[locMap[enriched[j].name]-1].adjustedPValue;
         double adj2=enriched[j].adjustedPValue;
         if (adj1==0)
           adj1=log(1e-323);
         else
           adj1=log(adj1);

         if (adj2==0)
           adj2=log(1e-323);
         else
           adj2=log(adj2);

         ontoFin.push_back(make_pair(enriched[j].name, adj1*adj2));
      }
   }

   sort(ontoFin.begin(), ontoFin.end(), compareOntoFin);

   getTopOntoPairs(ontoFin, parentChildFile, numbTopOnto, m.getName()); 
}

void MappingResults::getTopOntoPairs(vector<pair<string, double> > &ontoFin, string parentChild, int numbTopOnto, string domainName)
{
   ifstream in(parentChild.c_str());
   string line, s, s1;
   map<string, set<string> > parentChildMap;
   while (getline(in, line))
   {
      istringstream is(line);
      is>>s;
      while (is>>s1)
        parentChildMap[s].insert(s1);
   }

   vector<pair<string, double> >::iterator vpItr;
   for (vpItr=ontoFin.begin(); vpItr!=ontoFin.end(); ++vpItr)
   {
      if (topOnto[domainName].size()<numbTopOnto)
      {
         bool hasParent=false;
         if (topOnto[domainName].empty())
           topOnto[domainName].push_back(make_pair(vpItr->first, vpItr->second));
         else
         {
           for (int i=0; i<topOnto[domainName].size(); ++i)
           {
             if (parentChildMap[topOnto[domainName][i].first].find(vpItr->first)!=parentChildMap[topOnto[domainName][i].first].end()) 
             {
                hasParent=true;
                break;
             }
             else if (parentChildMap[vpItr->first].find(topOnto[domainName][i].first)!=parentChildMap[vpItr->first].end()) 
             {
                hasParent=true;
                break;
             }
           }
       
           if (!hasParent)
             topOnto[domainName].push_back(make_pair(vpItr->first, vpItr->second));
         }
      }
   }

}

void MappingResults::printTopOnto(ostream &out)
{
   map<string, vector<pair<string, double> > >::iterator mvpItr;
   for (mvpItr=topOnto.begin(); mvpItr!=topOnto.end(); ++mvpItr)
      for (int i=0; i<mvpItr->second.size(); ++i)
         out<<name<<"\t"<<mvpItr->first<<"\t"<<mvpItr->second[i].first<<"\t"<<mvpItr->second[i].second<<endl; 
}

void MappingResults::printTopOnto(ostream &out, string annotFile)
{
   ifstream in(annotFile.c_str());
   string id, annot;
   map<string, string> annotMap;
   while (getline(in, id, '\t'))
   {
      getline(in, annot);
      annotMap[id]=annot;
   } 
      
   map<string, vector<pair<string, double> > >::iterator mvpItr;
   for (mvpItr=topOnto.begin(); mvpItr!=topOnto.end(); ++mvpItr)
      for (int i=0; i<mvpItr->second.size(); ++i)
         out<<name<<"\t"<<mvpItr->first<<"\t"<<mvpItr->second[i].first<<"\t"<<annotMap[mvpItr->second[i].first]<<"\t"<<mvpItr->second[i].second<<endl;

}

void MappingResults::printTopOnto(ostream &out, string annotFile, string groupId)
{
   ifstream in(annotFile.c_str());
   string id, annot;
   map<string, string> annotMap;
   while (getline(in, id, '\t'))
   {
      getline(in, annot);
      annotMap[id]=annot;
   }

   map<string, vector<pair<string, double> > >::iterator mvpItr;
   for (mvpItr=topOnto.begin(); mvpItr!=topOnto.end(); ++mvpItr)
      for (int i=0; i<mvpItr->second.size(); ++i)
         out<<name<<"\t"
            <<mvpItr->first<<"\t"
            <<mvpItr->second[i].first<<"\t"
            <<annotMap[mvpItr->second[i].first]<<"\t"
            <<mvpItr->second[i].second<<"\t"
            <<groupId<<endl;
}


void MappingResults::printSummaryScore(ostream& out)
{
   map<string, double>::iterator mItr;
   for (mItr=summaryScore.begin(); mItr!=summaryScore.end(); ++mItr)
      if (mItr->second>0)
         out<<name<<"\t"<<mItr->first<<"\t"<<mItr->second<<endl;
}

void MappingResults::printSummaryScore(ostream& out, int percent)
{
   map<string, map<double, double> >::iterator mItr;
   map<double, double>::iterator midItr;
   if (percent == -1)
   {
      for (mItr=summaryScoreWithCutOff.begin(); mItr!=summaryScoreWithCutOff.end(); ++mItr)
      {
         for (midItr=mItr->second.begin(); midItr!=mItr->second.end(); ++midItr)
            if (midItr->second>0)
               out<<name<<"\t"
                  <<mItr->first<<"\t"
                  <<midItr->first<<"\t"
                  <<midItr->second<<endl;
      }
    }
    else
    {
      for (mItr=summaryScoreWithCutOff.begin(); mItr!=summaryScoreWithCutOff.end(); ++mItr)
      {  
         for (midItr=mItr->second.begin(); midItr!=mItr->second.end(); ++midItr)
            if (midItr->first==percent && midItr->second>0)
               out<<name<<"\t"<<mItr->first<<"\t"<<midItr->first<<"\t"<<midItr->second<<endl;
      }
    }
}

void MappingResults::addEnrichedWithAdjPValue(string ontoName, double pValue, double adjPValue)
{
   enrichedPair ep;
   ep.name = ontoName;
   ep.pValue = pValue;
   ep.adjustedPValue = adjPValue;
   enriched.push_back(ep);
}
